/* Generic FidoNet message interface
 */

#ifndef _msg_h
#define _msg_h
#include <time.h>

#include "array.h"
#include "strbuf.h"
#include "txtbuf.h"
#include "fidoaddr.h"

/* Enumerator for known types of message areas
 */

enum _msgbase
{
  AREA_SDM,                           /* Classic *.MSG */
  AREA_SQUISH,                        /* Squish area *.SQD/*.SQI */
  AREA_PKT22,                         /* PKT 2.2 (FSC-0045) */
  AREA_PKT2P,                         /* PKT 2+ (FSC-0039) */
  AREA_HMB,                           /* 'Hudson' message base */
  AREA_JAM,                           /* JAM message base */
  AREA_max
};


typedef long MSGNO;
typedef struct _msg MSG;
typedef struct _msgarea MSGAREA;

/* To make life easier for us when extending this application,
 * we use a similar technique to that used by C++, by defining
 * a set of virtual interface functions to the 'real' functions
 * used to handle the actual message base manipulation.
 * The 'vtable' is simply a struct that defines a set of
 * functions that is in fact the messagebase interface.
 */

typedef struct _mastats
{
  MSGNO first;                             /* First and last msg numbers */
  MSGNO last;
  long  total;                                     /* Number of messages */
  MSGNO pos;                          /* Current position (non-0 if set) */
} MASTATS;

/* Special message numbers for _msgvtable.msgopen() */
#define MSGNEW    ((MSGNO)0)
#define MSGFIRST  ((MSGNO)-1)
#define MSGLAST   ((MSGNO)-2)
#define MSGNEXT   ((MSGNO)-3)
#define MSGPREV   ((MSGNO)-4)

/* Values for 'mode' parm in _msgvtable.msgopen() */
#define MSGRDONLY   0           /* Read msg */
#define MSGWRONLY   1           /* Overwrite msg (& create) */
#define MSGRDWR     2           /* Read/write msg */

typedef struct _msgvtable
{                                      /* --- Area related functions --- */
  MSGAREA * (*areainit)(MSGAREA *ma, void *data);         /* Create area */
  void      (*areadestroy)(MSGAREA *ma);                 /* Destory area */
  int       (*areaopen)(MSGAREA *ma);                       /* Open area */
  int       (*areacreate)(MSGAREA *ma);         /* Create/overwrite area */
  int       (*areaclose)(MSGAREA *ma);                     /* Close area */
  int       (*arealock)(MSGAREA *ma, long timeout);         /* Lock area */
  int       (*areaunlock)(MSGAREA *ma);                   /* Unlock area */
  int       (*areakillmsg)(MSGAREA *ma, MSGNO which);      /* Kill a msg */
  MASTATS   (*areastat)(MSGAREA *ma);                 /* Area statistics */
                                        /* --- Msg related functions --- */
  MSGNO     (*msgopen)(MSG *msg, MSGNO which, int mode);     /* Open msg */
  void      (*msgclose)(MSG *msg);                          /* Close msg */
  int       (*msgread)(MSG *msg);                /* Read msg into memory */
  int       (*msgwrite)(MSG * msg);             /* Write msg from memory */
} MAVTABLE;


/* Useful defines to hide the vtable */
#define marea_init(a,d)   ((a)->vtbl.areainit(a,d))
#define marea_destroy(a)  ((a)->vtbl.areadestroy(a))
#define marea_open(a)     ((a)->vtbl.areaopen(a))
#define marea_create(a)   ((a)->vtbl.areacreate(a))
#define marea_close(a)    ((a)->vtbl.areaclose(a))
#define marea_lock(a,t)   ((a)->vtbl.arealock(a,t))
#define marea_unlock(a)   ((a)->vtbl.areaunlock(a))
#define marea_stat(a)     ((a)->vtbl.areastat(a))

#define msg_open(m,w,m)   (((m)->area)->vtbl.msgopen((m),w,m))
#define msg_close(m)      (((m)->area)->vtbl.msgclose(m))
#define msg_read(m)       (((m)->area)->vtbl.msgread(m))
#define msg_write(m)      (((m)->area)->vtbl.msgwrite(m))

#define marea_error(a)    ((a)->vtbl.error(a))

struct _msgarea
{
  MAVTABLE  *vtbl;                                    /* Area vtable */
  void      *data;                           /* Format specific data */
  Array     msgs;                             /* Array with msg data */
  MASTATS   stats;                                /* Area statistics */
  #define   MA_OPEN     0x0100    /* Area is open */
  #define   MA_ERROR    0x0200    /* Error in last call */
  #define   MA_ECHO     0x8000    /* Area is echomail */
  unsigned  aflags;                                    /* Area flags */
};

enum _msgstr                                /* Header string indices */
{
  M_FROM,                                               /* From name */
  M_TO,                                                   /* To name */
  M_SUBJECT,                                              /* Subject */
  M_DATE,                                       /* Text date written */
  M_MSGID,                                           /* Stored MSGID */
  M_REPLY,                                           /* Stored REPLY */
  M_max
};

/* Generic FidoNet message structure
 */

struct _msg
{
  MSGAREA         *area;                             /* Ptr to area */
  strbuf          str;                        /* Local strings heap */
  textbuf         txt;                              /* Message text */
  Array           replies;                               /* Replies */
  MSGNO           msgno;                          /* Message number */
  MSGNO           replyto;                      /* Original message */
  ZNNPD           orig,          /* Originating/destination address */
                  dest;
  time_t          written,            /* Times written/arrived-read */
                  arrived;
                                      /* --- FTS-1 attributes --- */
  #define   MSGPRIVATE  0x00000001L   /* Private */
  #define   MSGCRASH    0x00000002L   /* Crashmail */
  #define   MSGREAD     0x00000004L   /* Received */
  #define   MSGSENT     0x00000008L   /* Sent */
  #define   MSGFILE     0x00000010L   /* File attached */
  #define   MSGFWD      0x00000020L   /* In-transit */
  #define   MSGORPHAN   0x00000040L   /* Orphaned */
  #define   MSGKILL     0x00000080L   /* Kill after sending */
  #define   MSGLOCAL    0x00000100L   /* Written locally */
  #define   MSGHOLD     0x00000200L   /* Hold for pickup */
  #define   MSGXX2      0x00000400L   /* Unused */
  #define   MSGFRQ      0x00000800L   /* File request */
  #define   MSGRRQ      0x00001000L   /* Receipt requested */
  #define   MSGRCPT     0x00002000L   /* Is a receipt */
  #define   MSGARQ      0x00004000L   /* Audit request */
  #define   MSGURQ      0x00008000L   /* Update request */
                                      /* ---Extended attributes --- */
  #define   MSGIMM      0x00010000L   /* Immediate priority */
  #define   MSGDIRECT   0x00020000L   /* Direct */
  #define   MSGTFS      0x00040000L   /* Truncate file */
  #define   MSGKFS      0x00080000L   /* Erase file */
  #define   MSGRREQ     0x00100000L   /* Receipt request */
  #define   MSGCFM      0x00200000L   /* Confirmation request */
  #define   MSGFPU      0x00400000L   /* Force pickup */
  #define   MSGLOCKED   0x00800000L   /* Message is locked or read-only */
  #define   MSGXMAIL    0x01000000L   /* Compress */
  #define   MSGARCSENT  0x02000000L   /* Arc after sent */
  #define   MSGFAX      0x01000000L   /* Fax image */
  #define   MSGFAXCOV   0x02000000L   /* Fax cover sheet */
  #define   MSGFAXSIG   0x04000000L   /* Fax signature */
  #define   MSGFAXLET   0x08000000L   /* Fax letterhead */
  #define   MSGZGATE    0x10000000L   /* Zonegate routed */
  #define   MSGHROUTE   0x20000000L   /* Host/hub route */
  unsigned long   attribs;                    /* Message attributes */
  unsigned long   xattribs;               /* Application attributes */
  #define   MF_HDRS     0x00ff        /* # of header fields */
  #define   MF_OPEN     0x0100        /* Msg is open */
  #define   MF_WRITE    0x0200        /* Open allows write */
  #define   MF_INTL     0x0400        /* Forced ^aINTL */
  #define   MF_DOMAIN   0x0800        /* Forced ^aDOMAIN */
  #define   MF_NEW      0x8000        /* Msg is new */
  unsigned        mflags;                                  /* Flags */
  sofs_t          *hdrs;                            /* Text headers */
};


#endif

