
#include <string.h>
#include "array.h"
#include "mem.h"


/* Handle array (re)allocation
 */

static int
array_alloc(Array * A, unsigned tosize)
{
  if (tosize > A->asize)                    /* Only allocate if required */
  {
    unsigned memsize;
    char * newbase;
                        /* If just adding a new element, grow by a delta */
    tosize = (tosize == A->asize + 1) ? (tosize + (A->asize / 2)) : tosize;
    memsize = tosize * A->elsize;                /* Calc required memory */
    newbase = zrealloc("array_alloc", A->base, memsize);
    if (newbase == NULL)                                     /* Disaster */
      return -1;
    A->base = newbase;
    A->asize = tosize;
    return 1;
  }
  return 0;                                         /* Nothing allocated */
}


Array *
array_init(Array * A, unsigned elsize, unsigned isize)
{
  A->elsize = elsize;                     /* Size of individual elements */
  A->asize  = 0;                                /* Current size of array */
  A->aused  = 0;                         /* Zero number of elements used */
  A->base   = 0;
  array_alloc(A, isize);                   /* Allocate the initial array */
  return A;
}

/* Allocate a new array
 */

Array *
array_new(unsigned elsize, unsigned isize)
{
  return array_init(zmalloc("array_new", sizeof(Array)), elsize, isize);
}

void
array_destroy(Array * A)
{
  if (A)
    zfree("array_destroy", A->base);
}


/* Free an array
 */

void
array_delete(Array * A)
{
  array_destroy(A);
  zfree("array_delete", A);
}


/* Add new element to array
 */

unsigned
array_add(Array * A, unsigned where, unsigned elems)
{
  unsigned at = A->aused;

  if (elems)
  {
    A->aused += elems;
    if (array_alloc(A, A->aused) == -1)
      A->aused -= elems;
    else
    {
      if (where < at)          /* Need to move elements in the array up? */
        memmove(array_ptr(A, where + elems), array_ptr(A, where),
                (at - where) * A->elsize);
      else where = at;
      memset(array_ptr(A, where), 0, elems * A->elsize);
      return where;
    }
  }
  return NOELEMENT;
}


/* Delete elems from array
 */

unsigned
array_del(Array * A, unsigned where, unsigned elems)
{
  if (elems == NOELEMENT || (where + elems) >= A->aused)
    elems = (where > A->aused) ? 0 : A->aused - where;
  else
  { /* Need to move some elements down */
    unsigned top = where + elems;
    memmove(array_ptr(A, where), array_ptr(A, top),
              (A->aused - top) * A->elsize);
  }
  A->aused -= elems;
  return elems;   /* How many elements deleted */
}

/* Get pointer to an element
 */

void *
array_ptr(Array * A, unsigned idx)
{
  if (idx == NOELEMENT || idx >= A->aused)
    return NULL;
  return A->base + (A->elsize * idx);
}

/* Array iterator
 */

void
arrayIter_init(ArrayIter * I, Array * A)
{
  I->A = A;
  I->index = NOELEMENT;
}

/* Get an element */

void *
arrayIter_get(ArrayIter * I, int which)
{
  switch (which)
  {
    case PREV:
      if (I->index != NOELEMENT)
      {
        --I->index;
        break;
      }
      /* Fallthru */
    case LAST:
      I->index = (I->A)->aused - 1;
      break;
    case NEXT:
      if (I->index != NOELEMENT)
      {
        ++I->index;
        break;
      }
      /* Fallthru */
    case FIRST:
      I->index = 0;
      break;
  }
  if (I->index >= (I->A)->aused)    /* Auto reset at either end */
    I->index = NOELEMENT;
  return array_ptr(I->A, I->index);
}

void
array_reset(Array * A)
{
  A->aused = 0;
}

