/* Memory allocation functions
 * These are centralised for debugging and error handling
 * zmalloc() does not return unless allocation succeeds
 * zfree() does simplistic block checking to ensure we're freeing a valid block
 */

#include <stdlib.h>
#include <string.h>
#include "mem.h"
#include "log.h"

void *
zmalloc(char const *fn, size_t sz)
{
  void * p = NULL;
  if (sz)
  {
    unsigned * i = malloc(sz + sizeof(unsigned));
    if (i == NULL)
      fatal(EL_MEMERROR, "allocation error [%s] %u bytes", fn, sz);
    *i++ = MEM_MAGIC_BLOCK;
    p = i;
  }
  return p;
}

void *
zrealloc(char const * fn, void * p, size_t sz)
{
  if (p == NULL)
    p = zmalloc(fn, sz);
  else if (sz == 0)
  {
    zfree(fn, p);
    p = NULL;
  }
  else
  {
    unsigned * i = p;
    if (*--i != MEM_MAGIC_BLOCK)
      fatal(EL_MEMERROR, "realloc error [%s] corrupted heap?", fn);
    i = realloc(i, sz + sizeof(unsigned));
    if (i == NULL)
      p = NULL;
    else
    {
      *i++ = MEM_MAGIC_BLOCK;
      p = i;
    }
  }
  return p;
}

void
zfree(char const * fn, void * p)
{
  if (p)
  {
    unsigned * i = p;
    if (*--i != MEM_MAGIC_BLOCK)
      fatal(EL_MEMERROR, "free error [%s] corrupted heap?", fn);
    free(i);
  }
}

char *
zstrdup(char const * fn, char const * str)
{
  if (str == NULL)
    return NULL;
  else
  {
    size_t l = strlen(str) + 1;
    char * p = zmalloc(fn, l);
    memcpy(p, str, l);
    return p;
  }
}


