/* strbuf.c
 * Implements a simple strings buffer
 */

#include <stdlib.h>
#include <string.h>
#include "strbuf.h"
#include "mem.h"
#include "osdep.h"


#define sbptr(sb, ofs)     ((sb)->heap + (ofs))

strbuf *
sb_new(unsigned sz)
{
  return sb_init(zmalloc("sb_new", sizeof(strbuf)), sz);
}

strbuf *
sb_init(strbuf * sb, unsigned sz)
{
  if (sz <= 0)
    sz = STRBUFSZ;
  sb->soSize   = (sofs_t)0;
  sb->soOffset = (sofs_t)1;
  sb->heap     = 0;
  sb_realloc(sb, sz);
  return sb;
}

void
sb_destroy(strbuf * sb)
{
  if (sb)
    zfree("sb_destroy", sb->heap);
}

void
sb_delete(strbuf * sb)
{
  sb_destroy(sb);
  zfree("sb_delete", sb);
}

sofs_t
sb_salloc(strbuf * sb, char const * s, int slen)
{ /* Allocate a substring */
  static char const fn[] = "sb_salloc";
  sofs_t ofs = 0;

  if (s && *s && slen)
  {
    char * tmp = NULL;
    char * p;
    sofs_t len = slen + 1;                              /* Include NUL */
    long   top = (long)sb->soOffset + (long)len;

    if (top >= (long)sb->soSize)                /* Need to reallocate? */
    {
      if (top >= (sofs_t)-1)                /* Blown the maximum size? */
        return 0;                                /* Can't do much else */
                                 /* We may need to copy it temporarily */
      if (s > sb->heap && s < (sb->heap + sb->soOffset))
      {
        tmp = zmalloc(fn, len);
        memcpy(tmp, s, len);
        s = tmp;
      }                                      /* Make the buffer larger */
      if (!sb_realloc(sb, (unsigned)top))
        return 0;
    }
    p = sbptr(sb, ofs = sb->soOffset);    /* Get position of string */
    sb->soOffset += len;
    memcpy(p, s, len);                       /* Copy the new string in */
    p[len] = '\0';
    zfree(fn, tmp);                   /* Dispose of possible temporary */
  }
  return ofs;
}

sofs_t
sb_alloc(strbuf * sb, char const * s)
{ /* Allocate a standard string */
  return (sofs_t)(s == NULL ? 0 : sb_salloc(sb, s, strlen(s)));
}

void
sb_free(strbuf * sb, sofs_t ofs)
{
  if (ofs > 0 && ofs < sb->soOffset)
  {
    char * s = sbptr(sb, ofs);
    if (*(s-1) == '\0')
    {
      sofs_t at, len;
      sofs_t used = 1;
      *s=STRDELMK;                                 /* Delete this string */
                    /* Now, walk up the buffer and find the last string, */
                        /* adjusting the next offset downwards if we can */
      for (at = 1; at < sb->soOffset ; at += len)
      {
        char * p = sbptr(sb, at);
        len = (sofs_t)(strlen(p) + 1);
        if (*p != STRDELMK)             /* Track end of last used string */
          used = at + len;
      }  /* Set the last offset to one character past end of last string */
      sb->soOffset = used;
    }
  }
}

char const *
sb_string(strbuf * sb, sofs_t ofs)
{
  if (ofs < 1 || ofs >= sb->soOffset || ofs == STRNONE)
    ofs = 0;
  return sb->heap + ofs;
}

void
sb_reset(strbuf *sb)
{
  sb->soOffset = 1;
}

int
sb_realloc(strbuf * sb, int sz)
{
  sz += (STRBUFSZ-1);                  /* Increment in larger blocks */
  sz -= (sz % STRBUFSZ);
  if (sz > sb->soSize)
  {
    char * newheap;
    if ((newheap = zrealloc("sb_realloc", sb->heap, sz)) == NULL)
      return 0;
    sb->heap = newheap;
    *sb->heap = '\0';
    sb->soSize = sz;
  }
  return 1;
}

void
sb_pack(strbuf * sb, int (*reloc)(sofs_t mold, sofs_t mnew))
{
  static char const fn[] = "sb_pack";
  char * newheap = zmalloc(fn, sb->soOffset);
  sofs_t at, len, ofs = 1;

  *newheap = '\0';        /* Walk up buffer and discard deleted strings */
  for (at = 1; at < sb->soOffset ; at += len)
  {
    char * p = sbptr(sb, at);
    len = (sofs_t)(strlen(p) + 1);
    if (*p != STRDELMK && (!reloc || reloc(at, ofs)))
    {                         /* Drop the string if caller has lost it */
      memcpy(newheap + ofs , p, len);
      ofs += len;
    }
  }
  /* Move in the new buffer */
  zfree(fn, sb->heap);
  sb->heap = newheap;
  sb->soSize = sb->soOffset = ofs;
}

sofs_t
sb_offset(strbuf *sb, char *s, int icase)
{
  sofs_t  at, len;

  for (at = 1; at < sb->soOffset; at += len)
  {
    char * p = sbptr(sb, at);
    len = (sofs_t)(strlen(p) + 1);
    if ((icase ? stricmp : strcmp)(s, p)==0)
      return at;
  }
  return STRNONE;
}

