/* main.c
 * main module
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <stdarg.h>
#include <ctype.h>

#define NLMAINT

#include "nlmaint.h"
#include "osdep.h"
#include "version.h"
#include "filediff.h"
#include "log.h"
#include "cmdtpl.h"

static void usage(void);
static int update_description(NLCONFIG * nlc, char const * filesbbs, char const * tpl, char const * list, char const * fdif);

static char const * modes[] =
{
  "TEST", "PROCESS"
};

static void
usage(void)
{
  printf("usage:\n"
         "%s [-p|-t|-s] [-m=<list>] [-n=<name>] [-f] [-h|-?] [<ctlfile[.CTL]>]\n"
         "  -p[rocess]      Force process mode (make list & submit)\n"
         "  -t[est]         Force test mode (scan & check submittals only)\n"
         "  -s[uppress]     Force process mode, but skip archive & submit\n"
         "  -m[erge]=<list> Merge resulting segment with <list>\n"
         "  -f[orce]        Force submit, even if segment has not changed\n"
         "  -n[ame]=<name>  Override name of network to insert into list\n"
         "  -h[elp] or -?   Print this help message\n"
#ifdef DOSISH
         "  Use of '/' as switch character is optional, case of switches is ignored\n\n"
#endif
	 ,
         progname);
}

int
main (int argc, char *argv[])
{
  int rc = EL_CTLERROR,
      carg = 0,
      force_process = 0,
      force_test = 0,
      no_submit = 0,
      force_submit = 0;
  sofs_t * so;
  ArrayIter I;
  char const *config_file = NULL;
  char const *merge_file = NULL;
  char const *net_name = NULL;
  struct tm * T;
  NLCONFIG * nlc =0;
  long days;

  printf("%s; FidoNet Nodelist Maintenance Tool; Ver %s\n"
         "by " AUTHOR "\n"
         "Compiled for " HOSTOS " on " __DATE__ " at " __TIME__ "\n\n",
         progname, version);

  tzset();
  while (++carg < argc)
  {
    char * argp = argv[carg];

#ifdef DOSISH
    if (*argp == '-' || *argp == '/')  /* Command switch */
#else
    if (*argp == '-')		       /* Command switch */
#endif
    {
      char const * arg = strtok(++argp, "=:");
      switch (cmdlineverb(arg))
      {
      case ARG_DEFAULT:   /* Shouldn't happen */
        break;

      case ARG_PROCESS:   /* Force processing */
        if (force_test || no_submit)
          goto notargs;
        ++force_process;
        break;

      case ARG_TEST:      /* Force test */
        if (force_process || no_submit)
          goto notargs;
        ++force_test;
        break;

      case ARG_SUPPRESS:  /* Suppress submissions */
        if (!force_process && !force_test)
        {
          force_process = no_submit = 1;
          break;
        }
      notargs:
        fatal(EL_CMDERROR, "Only one of /T[est] /P[rocess] or /S[ubmit] may be specified");
        break;

      case ARG_MERGE:     /* Merge list */
        merge_file = strtok(NULL, "");
        if (merge_file == NULL && (merge_file = argv[++carg]) == NULL)
          fatal(EL_CMDERROR, "/M[erge] specified with no list to merge");
        break;

      case ARG_NAME:    /* Network name */
        net_name = strtok(NULL, "");
        if (net_name == NULL && (net_name = argv[++carg]) == NULL)
          fatal(EL_CMDERROR, "/N[ame] specified with no network name");
        break;

      case ARG_FORCE:     /* Force submit, even if no change */
        force_submit = 1;
        break;

      case ARG_HELP: case ARG_ALTHELP:
        usage();
        fatal(0, NULL);

      default:
        usage();
        fatal(EL_CMDERROR, "Unknown command line switch /%s", arg);
        break;
      }
    }
    else if (config_file == NULL)
      config_file = argp;
    else
    {
      usage();
      fatal(EL_CMDERROR, "Unknown word on command line \"%s\"", argp);
    }
  }

  /* Use the default config file if none was specified
   */
  if (config_file == NULL)
    config_file = DFLTCFG;

  /* Read the configuration file
   */

  configure(&nlc, config_file);

  /* Adjust for command line overrides
   */

  if (net_name != NULL)
    nlc->str[V_NAME] = sb_alloc(&nlc->strings, net_name);
  if (merge_file != NULL)
    nlc->str[V_MERGE] = sb_alloc(&nlc->strings, merge_file);

  /* Open our log file
   */

  openlog(nlstring(nlc, V_LOGFILE), nlc->logformat, nlc->loglevel);

  /* Now, determine our target and which operating mode we want
   */

  nlc->now = time(NULL);
  T = localtime(&nlc->now);
  if (force_test)
    nlc->mode = MODE_TEST;
  else if (force_process)
    nlc->mode = MODE_PROCESS;
  /* Test processing mode unless on processing day of week */
  else nlc->mode = (T->tm_wday == nlc->process) ? MODE_PROCESS : MODE_TEST;
  if (nlc->process == V_NEVER)
    nlc->process = nlc->publish;

  /* Calculate number of days forward that next processing date falls */
  days = (T->tm_wday > nlc->process)  ? (7 - (T->tm_wday - nlc->process))
                                      : nlc->process - T->tm_wday;
  /* Now, calculate when the subsequent publishing date is after that */
  days += (nlc->publish < nlc->process) ? (7 - (nlc->process - nlc->publish))
                                        : (nlc->publish - nlc->process);
  /* Then, let's find out the date, and therefore the julian day to process */
  nlc->publish_time = nlc->now + (days * SECSPERDAY);
  T = localtime(&nlc->publish_time);
  nlc->target = (T->tm_yday + 1); /* tm_yday is days after 1st Jan */
  nlc->tyear = T->tm_year + 1900;
  logit(LOG_EVENT, "Running in %s mode for julian day %d", modes[nlc->mode], nlc->target);

  /* First, we search our inbound directories for likely input files */
  arrayIter_init(&I, &nlc->Inbound);
  while ((so = arrayIter_get(&I, NEXT)) != NULL)
    move_inbound(nlc, nlstr(nlc, *so));

  /* Ok, that finishes with things done in both modes
   * Now, let's see if there's anything to process...
   */

  if (nlc->mode != MODE_PROCESS)
    rc = test_segment(nlc) ? EL_TESTERRORS : EL_TESTOK;
  else
  {
    int dodiff;
    char targ[_MAX_PATH] = "",  /* Target nodelist */
         mlst[_MAX_PATH] = "",  /* List to merge */
         fdif[_MAX_PATH] = "",  /* Difference file */
         ddel[_MAX_PATH] = "";  /* Diff to delete */

    move_updates(nlc);    /* Move list updates, apply difference files */

    if (!nlsnull(nlc,V_BEFORE))
      exec_cmd(nlc, nlstring(nlc, V_BEFORE), NULL, NULL, NULL, NULL);

    if ((dodiff = make_segment(nlc, targ)) != -1)
    {
      char dest[_MAX_PATH];
      char *submit = targ;        /* File to submit */

      strcpy(mlst, targ);         /* Save this for merge */
      logit(LOG_EVENT, "New nodelist \"%s\" made successfully", targ);
      if (dodiff)
      {
        int d = make_diff(nlc, targ, fdif, force_submit);
        if (d==0)
        {
          logit(LOG_EVENT, "Difference file \"%s\" made successfully", fdif);
          submit = fdif;    /* Submit the diff file */
        }
        else if (d == 1)
        {
          logit(LOG_EVENT, "Not submitting unchanged file");
          if (!no_submit)   /* We might be testing here */
            remove(targ);   /* Otherwise we can junk this file */
          no_submit = 1;    /* Force no submittal either way */
        }
        else /* if (d == -1) */
          dodiff = 0;
      }

      if (no_submit)
        logit(LOG_EVENT, "Skipping archive & submit");
      else
      {
        int nodearc = 0;
        int diffarc = 0;
        static char const filesbbs[] = "files.bbs";

        if (!nlsnull(nlc, V_ARCLIST) && !nlsnull(nlc, V_ARC))
        {
          if (archive_list(nlc, targ, nlstring(nlc, V_ARCLIST), nlstring(nlc, V_ARC), nlstring(nlc, V_NLPFX), 0) == 0)
            ++nodearc;
          else logit(LOG_WARN, "Archiver failed - using raw list instead");

          if (!nodearc && copylist(targ, nlstring(nlc, V_ARCLIST)))
            ++nodearc;    /* Now means succeded, to update FILES.BBS */
        }

        if (nlsnull(nlc, V_DIFFARC))
          dodiff = 0;
        else if (dodiff)
        {                            /* Only do this if we have generated a diff */
          strcpy(ddel, fdif);
          if (archive_list(nlc, fdif, nlstring(nlc, V_ARCDIFF), nlstring(nlc, V_DIFFARC), nlstring(nlc, V_NDPFX), 0) == 0)
            ++diffarc;
          else logit(LOG_WARN, "Archiver failed - using raw diff instead");

          if (!diffarc && copylist(fdif, nlstring(nlc, V_ARCDIFF)))
            ++diffarc;    /* Now means succeded, to update FILES.BBS */
        }

        if (nodearc && !nlsnull(nlc, V_LISTDESC))
          update_description(nlc, build_path(dest, get_path(NULL, targ), filesbbs, NULL), nlstring(nlc, V_LISTDESC), targ, fdif);

        if (diffarc && !nlsnull(nlc, V_DIFFDESC))
          update_description(nlc, build_path(dest, get_path(NULL, fdif), filesbbs, NULL), nlstring(nlc, V_DIFFDESC), targ, fdif);

        if (nlsnull(nlc,V_EXECSUBMIT))
        {
          if (nlc->submitaddr.net)
            logit(LOG_MARK, "No EXECSUBMIT command, can't submit \"%s\" to %s", submit, fidofmt(NULL, &nlc->submitaddr, NULL));
        }
        else exec_cmd(nlc, nlstring(nlc, V_EXECSUBMIT), targ, fdif, submit, NULL);

        if (!nlsnull(nlc,V_AFTER))
          exec_cmd(nlc, nlstring(nlc, V_AFTER), targ, fdif, submit, NULL);

        /* It is safe to delete the diff file now */
        if (*ddel)
          remove(ddel);

        if (!nlsnull(nlc,V_MERGE))
        {
          if (!find_latest(targ, nlstring(nlc,V_MERGE), F_LIST, nlc->publish_time))
            logit(LOG_WARN, "Can't find any nodelist \"%s\" for merge", nlstring(nlc,V_MERGE));
          else
          {
            strcpy(fdif, targ);
            add_extn(fdif, 999);    /* Use <listname>.999 as target */
            logit(LOG_PROGRESS, "Merging \"%s\" with \"%s\" to produce \"%s\"", last_component(targ), last_component(mlst), fdif);
            if (merge_list(fdif, targ, mlst, nlc->myaddr) == 0)
              logit(LOG_RESULT, "Merge successful. Output file is \"%s\"", fdif);
            else logit(LOG_ERROR, "Merge failed");
          }
        }

        /*
         * Finally, let's clean up everything and keep only the latest
         * copy of input and output files in the MASTER and OUTPUT dirs
         */

        if (nlc->cleanup)
        {
          char const * file;
          subfile * sf;
          ArrayIter iter;

          /* First, segments in the MASTER directory */
          arrayIter_init(&iter, &nlc->Files);
          while ((sf = arrayIter_get(&iter, NEXT)) != NULL)
          {
            file = nlstr(nlc, sf->name);
            build_path(dest, nlstring(nlc, V_MASTER), file, NULL);
            if (find_latest(dest, dest, F_LIST, nlc->publish_time) > 1)
              cleanup_lists(nlstring(nlc, V_MASTER), file, dest, nlc->publish_time);
          }

          /* Then output lists */
          file = nlstring(nlc, V_OUTFILE);
          build_path(dest, nlstring(nlc, V_OUTPATH), file, NULL);
          if (find_latest(dest, dest, F_LIST, nlc->publish_time) > 1)
            cleanup_lists(nlstring(nlc, V_OUTPATH), file, dest, nlc->publish_time);
        }
      }
    }
    rc = (nlc->errcount) ? EL_PROCESSERRORS : EL_PROCESSOK;
  }

  send_feedback(nlc, no_submit);

  logit(LOG_MARK, "NLMAINT exiting with status %d", rc);
  openlog(NULL, -1, 0);
  return rc;
}

/* This is an enumerator for the template array below
 * This makes the array easier to manipulate, since all you
 * have to do is keep the two in sync rather than having to
 * modify code.
 */

enum
{
  t_DAYNUM, t_SDAYNUM, t_OLDNUM, t_SOLDNUM,
  t_MASTER, t_OUTPATH, t_MSGBASE, t_NETNAME,
  t_MYADDR, t_SUBMIT,
  t_NLPATH, t_NLNAME, t_NLBASE, t_NLEXT,
  t_NDPATH, t_NDNAME, t_NDBASE, t_NDEXT,
  t_ARCLIST, t_ARCDIFF, t_NLCOMP, t_NDCOMP, t_NLCMPL, t_NDCMPL,
  t_NLSIZE, t_NDSIZE, t_NLDATE,
  t_USER, t_USERNAME, t_USERDIR, t_USERBASE, t_USEREXT,
  t_SFADDR, t_SFNAME, t_SFNUM, t_SFPWD, t_NOTIFY
};

/* Common substitution template for cmds and FILES.BBS updating
 */
static cmdtpl tplsubs[] =
{
  { 'y',  TC_STD,   0,  NULL, NULL  },  /* t_DAYNUM   Day number */
  { 'Y',  TC_STD,   0,  NULL, NULL  },  /* t_SDAYNUM  Day number left truncated to 2 chrs */
  { 'p',  TC_STD,   0,  NULL, NULL  },  /* t_OLDNUM   Day number for previous week */
  { 'P',  TC_STD,   0,  NULL, NULL  },  /* t_SOLDNUM  Previous weeks number left truncated */
  { 'm',  TC_STD,   0,  NULL, NULL  },  /* t_MASTER   Master directory */
  { 'o',  TC_STD,   0,  NULL, NULL  },  /* t_OUTPATH  Output directory */
  { 'M',  TC_STD,   0,  NULL, NULL  },  /* t_MSGBASE  Messagebase */
  { 'e',  TC_STD,   0,  NULL, NULL  },  /* t_NETNAME  Network name */
  { 'O',  TC_STD,   0,  NULL, NULL  },  /* t_MYADDR   My network address */
  { 't',  TC_STD,   0,  NULL, NULL  },  /* t_SUBMIT   Submit address */
  { 'n',  TC_STD,   0,  NULL, NULL  },  /* t_NLPATH   Nodelist name (maybe archived) */
  { 'N',  TC_NAME,  0,  NULL, NULL  },  /* t_NLNAME   Name of nodelist */
  { 'B',  TC_BASE,  0,  NULL, NULL  },  /* t_NLBASE   Base name of nodelist */
  { 'E',  TC_EXT,   0,  NULL, NULL  },  /* t_NLEXT    Nodelist file extension */
  { 'd',  TC_STD,   0,  NULL, NULL  },  /* t_NDPATH   Nodediff name */
  { 'D',  TC_NAME,  0,  NULL, NULL  },  /* t_NDNAME   Name of nodediff */
  { 'b',  TC_BASE,  0,  NULL, NULL  },  /* t_NDBASE   Base name of nodediff */
  { 'e',  TC_EXT,   0,  NULL, NULL  },  /* t_NDEXT    Nodediff file extension */
  { 'S',  TC_STD,   0,  NULL, NULL  },  /* t_ARCLIST  ARCList sub-directory */
  { 's',  TC_STD,   0,  NULL, NULL  },  /* t_ARCDIFF  ARCDiff sub-directory */
  { 'A',  TC_STD,   0,  NULL, NULL  },  /* t_NLCOMP   Nodelist compression method */
  { 'a',  TC_STD,   0,  NULL, NULL  },  /* t_NDCOMP   Nodediff compression method */
  { 'L',  TC_STD,   0,  NULL, NULL  },  /* t_NLCMPL   Nodelist compression letter */
  { 'l',  TC_STD,   0,  NULL, NULL  },  /* t_NDCMPL   Nodediff compression letter */
  { 'I',  TC_STD,   0,  NULL, NULL  },  /* t_NLSIZE   Size of nodelist (bytes) */
  { 'i',  TC_STD,   0,  NULL, NULL  },  /* t_NDSIZE   Size of nodediff (bytes) */
  { 'c',  TC_STD,   0,  NULL, NULL  },  /* t_NLDATE   Nodelist publish date DD-Mmm-YYYY */
  { 'u',  TC_STD,   0,  NULL, NULL  },  /* t_USER     User specified string */
  { 'U',  TC_NAME,  0,  NULL, NULL  },  /* t_USERNAME Same - name only */
  { 'Y',  TC_DIR,   0,  NULL, NULL  },  /* t_USERDIR  Same - dir only */
  { 'z',  TC_BASE,  0,  NULL, NULL  },  /* t_USERBASE Same - basename only */
  { 'Z',  TC_EXT,   0,  NULL, NULL  },  /* t_USEREXT  Same - extension only */
  { 'f',  TC_STD,   0,  NULL, NULL  },  /* t_SFADDR   Subfile address */
  { 'F',  TC_STD,   0,  NULL, NULL  },  /* t_SFNAME   Subfile name */
  { 'g',  TC_STD,   0,  NULL, NULL  },  /* t_SFNUM    Subfile number */
  { 'G',  TC_STD,   0,  NULL, NULL  },  /* t_SFPWD    Subfile password */
  { 'T',  TC_STD,   0,  NULL, NULL  },  /* t_NOTIFY   Notify priority parameters */
  {   0,  TC_LAST,  0,  NULL, NULL  }
};



static void
make_template(NLCONFIG * nlc, char const * list, char const * fdif, char const * user, subfile *sf)
{
  time_t lastweek = nlc->publish_time - (SECSPERDAY*7);
  struct tm * T = localtime(&lastweek);
  struct stat st;
  unsigned i;
  static char daynum[6],      /* current day number */
              oldnum[6],      /* Last week's day number */
              myaddr[32],     /* My address */
              toaddr[32],     /* Submit address */
              sfaddr[32],     /* Subfile address */
              sfnum[16],      /* Subfile number */
              nlsize[16],     /* Nodelist size */
              ndsize[16],     /* Nodediff size */
              nldate[16],     /* Nodelist date */
              notify[64];     /* Notify attributes */

  static char const mns[] = "JanFebMarAprMayJunJulAugSepOctNovDec";

  sprintf(daynum, "%03d", nlc->target);
  sprintf(oldnum, "%03d", T->tm_yday + 1);
  fidofmt(myaddr, &nlc->myaddr, NULL);
  fidofmt(toaddr, &nlc->submitaddr, NULL);
  fidofmt(sfaddr, sf ? &sf->addr : &nlc->myaddr, NULL);
  sprintf(sfnum, "%u", sf ? sf->number : nlc->number);
  if (list && *list && stat(list, &st) == 0)
    sprintf(nlsize, "%ld", (long)st.st_size);
  else strcpy(nlsize, "0");
  if (fdif && *fdif && stat(fdif, &st) == 0)
    sprintf(ndsize, "%ld", (long)st.st_size);
  else strcpy(ndsize, "0");
  T = localtime(&nlc->now);
  sprintf(nldate, "%02d-%3.3s-%04d", T->tm_mday, mns+(T->tm_mon*3), T->tm_year+1900);
  i = (sf ? sf->notify : nlc->notify) & ~(NF_NOBODY|NF_RECEIPT|NF_ERRORS|NF_ALL);
  if (~(i & NF_XFLAGS))
    i &= ~(NF_IMM|NF_DIRECT);
  notify_str(notify, i);

  for (i = 0; tplsubs[i].flags != TC_LAST; i++)
  {
    switch (i)
    {
    case t_DAYNUM:
      tplsubs[i].arg = daynum;
      break;
    case t_SDAYNUM:
      tplsubs[i].arg = daynum+1;
      break;
    case t_OLDNUM:
      tplsubs[i].arg = oldnum;
      break;
    case t_SOLDNUM:
      tplsubs[i].arg = oldnum+1;
      break;
    case t_MASTER:
      tplsubs[i].arg = nlstring(nlc,V_MASTER);
      break;
    case t_OUTPATH:
      tplsubs[i].arg = nlstring(nlc,V_OUTPATH);
      break;
    case t_MSGBASE:
      tplsubs[i].arg = nlstring(nlc,V_MESSAGES);
      break;
    case t_NETNAME:
      tplsubs[i].arg = nlstring(nlc,V_NAME);
      break;
    case t_MYADDR:
      tplsubs[i].arg = myaddr;
      break;
    case t_SUBMIT:
      tplsubs[i].arg = toaddr;
      break;
    case t_NLPATH: case t_NLNAME: case t_NLBASE: case t_NLEXT:
      tplsubs[i].arg = list;
      break;
    case t_NDPATH: case t_NDNAME: case t_NDBASE: case t_NDEXT:
      tplsubs[i].arg = fdif;
      break;
    case t_ARCLIST:
      tplsubs[i].arg = nlstring(nlc,V_ARCLIST);
      break;
    case t_ARCDIFF:
      tplsubs[i].arg = nlstring(nlc,V_ARCDIFF);
      break;
    case t_NLCOMP:
      tplsubs[i].arg = nlstring(nlc,V_ARC);
      break;
    case t_NDCOMP:
      tplsubs[i].arg = nlstring(nlc,V_DIFFARC);
      break;
    case t_NLCMPL:
      tplsubs[i].arg = nlstring(nlc,V_NLPFX);
      break;
    case t_NDCMPL:
      tplsubs[i].arg = nlstring(nlc,V_NDPFX);
      break;
    case t_NLSIZE:
      tplsubs[i].arg = nlsize;
      break;
    case t_NDSIZE:
      tplsubs[i].arg = ndsize;
      break;
    case t_NLDATE:
      tplsubs[i].arg = nldate;
      break;
    case t_USER: case t_USERNAME: case t_USERDIR: case t_USERBASE: case t_USEREXT:
      tplsubs[i].arg = user;
      break;
    case t_SFADDR:
      tplsubs[i].arg = sfaddr;
      break;
    case t_SFNAME:
      tplsubs[i].arg = sf ? nlstr(nlc, sf->name) : nlstring(nlc,V_OUTFILE);
      break;
    case t_SFNUM:
      tplsubs[i].arg = sfnum;
      break;
    case t_SFPWD:
      tplsubs[i].arg = sf ? nlstr(nlc, sf->password) : nlstring(nlc,V_PASSWORD);
      break;
    case t_NOTIFY:
      tplsubs[i].arg = notify;
      break;
    }
  }
}

/* Execute a pre- or post- processing command template
 */

int
exec_cmd(NLCONFIG * nlc, char const * tpl, char const * list, char const * fdif, char const * user, subfile * sf)
{
  make_template(nlc, list, fdif, user, sf);
  return tpl_exec(tpl, tplsubs);
}

static int
update_description(NLCONFIG * nlc, char const * filesbbs, char const * tpl, char const * list, char const * fdif)
{
  int rc = 0;
  char * tmp;
  make_template(nlc, list, fdif, filesbbs, NULL);
  if ((tmp = tpl_make(tpl, tplsubs, 2048)) != NULL)
  {
    FILE * fp = fopen(filesbbs, "a+");

    if (fp==NULL)
      logit(LOG_WARN, "Can't update description in \"%s\"", filesbbs);
    else
    {
      fprintf(fp, "%s\n", tmp);
      fclose(fp);
      rc = 1;
    }
    tpl_dealloc(tmp, tplsubs);
  }
  return rc;
}

int
copylist(char * src, char const * destdir)
{
  char tmp[_MAX_PATH];

  build_path(tmp, destdir, last_component(src), NULL);
  if (stricmp(tmp, src) == 0)
    return 1;                   /* Pretend success if we don't need to copy it */
  if (fcopy(tmp, src, 0) == 0)
  {
    strcpy(src, tmp);           /* Update source buffer */
    return 1;
  }
  logit(LOG_WARN, "File copy \"%s\" => \"%s\" failed", src, destdir);
  return 0; /* Failure */
}

static char *
fb_name(NLCONFIG * nlc, subfile * sf)
{
  char const * base = sf ? nlstr(nlc,sf->name) : nlstring(nlc,V_OUTFILE);
  char * p = build_path(NULL, nlstring(nlc,V_MASTER), base, NULL);
  strcpy(path_ext(p), ".fbm");
  return p;
}

/* This outputs data to a temp file used for feedback
 */

int
fb_printf(NLCONFIG * nlc, subfile * sf, char const * fmt, ...)
{
  int rc;
  va_list argp;
  FILE * fbfp = sf ? sf->fbfp : nlc->fbfp;

  /* Global messsge override */
  if (((sf == NULL) ? nlc->notify : sf->notify) & NF_NOBODY)
    return 0;

  /* We must open the file if it is not already open
   */
  if (fbfp == NULL)
  {
    char ldate[64];
    struct tm * t;
    char * p = fb_name(nlc, sf);

    remove(p);
    if ((fbfp = fopen(p, "a")) == NULL)
      return 0;
    t = localtime(&nlc->now);
    strftime(ldate, sizeof ldate - 1, "%H:%M on %A, %d %B %Y", t);
    fprintf(fbfp, "%s %s in %s mode at %s\n", progname, version, modes[nlc->mode], ldate);
    if (sf)
      sf->fbfp = fbfp;
    else nlc->fbfp = fbfp;
  }
  va_start(argp, fmt);
  rc = vfprintf(fbfp, fmt, argp);
  va_end(argp);
  return rc;
}

static FILE *
feedback(NLCONFIG * nlc, FILE * fbfp, subfile * sf, int nosend)
{
  if (fbfp)
  {
    char fbfile[_MAX_PATH];

    static int done = 0;

    fclose(fbfp);                       /* Close the feedback file */
    strcpy(fbfile, fb_name(nlc, sf));              /* Get its name */
    if (nosend)
      remove(fbfile);
    else
    {
      if (!nlsnull(nlc,V_EXECPOST))
      {
        if (exec_cmd(nlc, nlstring(nlc,V_EXECPOST), NULL, NULL, fbfile, sf) == 0)
          remove(fbfile);      /* Posted, so we can safely remove it */
        else logit(LOG_RESULT, "Feedback file \"%s\" retained", fbfile);
      }
      else if (!done++)                     /* Only log warning once */
        logit(LOG_RESULT, "No EXECPOST cmd - feedback mail left in \"%s\"", build_path(NULL, nlstring(nlc,V_MASTER), "*.fbm", NULL));
    }
  }
  return NULL;
}

void
send_feedback(NLCONFIG * nlc, int nosend)
{
  subfile * sf;
  ArrayIter iter;

  nlc->fbfp = feedback(nlc, nlc->fbfp, NULL, 0);
  arrayIter_init(&iter, &nlc->Files);
  while ((sf = arrayIter_get(&iter, NEXT)) != NULL)
    sf->fbfp = feedback(nlc, sf->fbfp, sf, nosend);
}

