/*
 * Scans a nodelist, check for parsing errors
 */

#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <time.h>
#include <ctype.h>
#include "wfile.h"
#include "fidoaddr.h"
#define SCANLIST
#include "version.h"
#include "osdep.h"
#include "crc16.h"
#include "log.h"

enum
{
  XIT_OK      =0,
  XIT_ERRFILE =1,
  XIT_ERRFMT  =2,
  XIT_ERRCRC  =3,
  XIT_UNKNOWN =4
};

int
main (int argc, char *argv[])
{
  int rc = XIT_ERRFILE;

  printf("%s; FidoNet Nodelist File Checker; Ver %s\n"
         "by " AUTHOR "\n"
         "Compiled for " HOSTOS " on " __DATE__ " at " __TIME__ "\n\n",
         progname, version);

  tzset();
  if (argc < 2)
    fprintf(stderr, "usage: %s <nodelist_filename>\n", argv[0]);
  else
  {
    WFILE * w = wfile_new(argv[1]);
    if (!wfile_open(w))
      logit(LOG_ERROR, "Can't open %s: %s", argv[1], strerror(errno));
    else
    {
      clock_t taken, start = clock();
      unsigned lines = 0;
      unsigned nodes = 0;
      unsigned coords = 0;
      unsigned admins = 0;
      unsigned unknown = 0;
      unsigned comments = 0;
      unsigned short crcval = 0;
      unsigned short listcrc = (unsigned short)-1;
      char const * p;
      int len;

      rc = XIT_OK;
      logit(LOG_PROGRESS, "Scanning nodelist: %s", argv[1]);
      while ((p = wfile_getline(w, &len)) != NULL)
      {
        if (lines++ > 0)
          crcval = update_crc16((unsigned char const *)p, len, crcval);
        else
        {
          char const * q = memchr(p, ':', len);
          if (q == NULL || (q += 2, !isdigit(*q)))
          {
            logit(LOG_FATAL, "Invalid file format (no valid CRC in file)");
            rc = XIT_ERRFMT;
            break;
          }
          listcrc = (unsigned short)atol(q);
        }
        if (!wfile_isvalid(w))
        {
          logit(LOG_ERROR, "%s(%s): Invalid line format", argv[1], lines);
          ++unknown;
        }
        else if (wfile_iscomment(w))
          ++comments;   /* Don't parse */
        else
        {
          if (wfile_isadmin(w))
          {
            ++admins;
            if (wfile_iscoord(w))
            {
              ++coords;
              if (wfile_iszc(w))
                printf("Zone %3u  Reg %3s  \r", wfile_zone(w), "-");
              else if (wfile_isrc(w))
                printf("Zone %3u  Reg %3u  \r", wfile_zone(w), wfile_net(w));
              fflush(stdout);
            }
          }
          else ++nodes;
          /* At this point, we have an apparently valid node record
           * We now need to parse it fully to check the various fields.
           */
        }
      }
      taken = clock() - start;
      /* Print out some stats for benchmarking */
      logit(LOG_MARK, "%u line(s) scanned in %lu.%02lu seconds.", lines, taken / CLOCKS_PER_SEC, ((taken % CLOCKS_PER_SEC) * 100) / CLOCKS_PER_SEC);
      if (rc == XIT_OK)
      {
        if (listcrc == crcval)
        {
          logit(LOG_RESULT, "CRC-16 (%u) is OK", listcrc);
          if (unknown)
            rc = XIT_UNKNOWN;
        }
        else
        {
          logit(LOG_ERROR, "CRC mismatch - got %u, needed %u", crcval, listcrc);
          rc = XIT_ERRCRC;
        }
        logit(LOG_RESULT, "    Comment lines : %6u", comments);
        logit(LOG_RESULT, " Coorinator nodes : %6u", coords);
        logit(LOG_RESULT, "Other Admin nodes : %6u", admins - coords);
        logit(LOG_RESULT, "  Non-admin nodes : %6u", nodes);
        logit(unknown ? LOG_ERROR : LOG_RESULT, "    Invalid lines : %6u", unknown);
      }
      wfile_close(w);
    }
  }
  logit(LOG_PROGRESS, "Exiting with status %u", rc);
  return 0;
}

